import sys
import threading
from typing import Any

from ._types import WebsocketMessage
from .http import HTTP1Settings, HTTP2Settings

__version__: str
BUILD_GIL: bool

class RSGIHeaders:
    def __contains__(self, key: str) -> bool: ...
    def keys(self) -> list[str]: ...
    def values(self) -> list[str]: ...
    def items(self) -> list[tuple[str, str]]: ...
    def get(self, key: str, default: Any = None) -> Any: ...

class RSGIHTTPStreamTransport:
    async def send_bytes(self, data: bytes) -> None: ...
    async def send_str(self, data: str) -> None: ...

class RSGIHTTPProtocol:
    async def __call__(self) -> bytes: ...
    def __aiter__(self) -> Any: ...
    async def client_disconnect(self) -> None: ...
    def response_empty(self, status: int, headers: list[tuple[str, str]]) -> None: ...
    def response_str(self, status: int, headers: list[tuple[str, str]], body: str) -> None: ...
    def response_bytes(self, status: int, headers: list[tuple[str, str]], body: bytes) -> None: ...
    def response_file(self, status: int, headers: list[tuple[str, str]], file: str) -> None: ...
    def response_file_range(
        self, status: int, headers: list[tuple[str, str]], file: str, start: int, end: int
    ) -> None: ...
    def response_stream(self, status: int, headers: list[tuple[str, str]]) -> RSGIHTTPStreamTransport: ...

class RSGIWebsocketTransport:
    async def receive(self) -> WebsocketMessage: ...
    async def send_bytes(self, data: bytes) -> None: ...
    async def send_str(self, data: str) -> None: ...

class RSGIWebsocketProtocol:
    async def accept(self) -> RSGIWebsocketTransport: ...
    def close(self, status: int | None) -> tuple[int, bool]: ...

class RSGIProtocolError(RuntimeError): ...
class RSGIProtocolClosed(RuntimeError): ...

class WSGIScope:
    def to_environ(self, environ: dict[str, Any]) -> dict[str, Any]: ...

class WorkerSignal:
    def __init__(self) -> None: ...
    def set(self) -> None: ...

class WorkerSignalSync:
    qs: threading.Event

    def __init__(self) -> None: ...
    def set(self) -> None: ...

class ASGIWorker:
    def __new__(
        cls,
        worker_id: int,
        sock: Any,
        threads: int,
        blocking_threads: int,
        py_threads: int,
        py_threads_idle_timeout: int,
        backpressure: int,
        http_mode: str,
        http1_opts: HTTP1Settings | None,
        http2_opts: HTTP2Settings | None,
        websockets_enabled: bool,
        static_files: tuple[str, str, str | None, str | None] | None,
        ssl_enabled: bool,
        ssl_cert: str | None,
        ssl_key: str | None,
        ssl_key_password: str | None,
        ssl_protocol_min: str,
        ssl_ca: str | None,
        ssl_crl: list[str],
        ssl_client_verify: bool,
    ) -> ASGIWorker: ...

class WSGIWorker:
    def __new__(
        cls,
        worker_id: int,
        sock: Any,
        threads: int,
        blocking_threads: int,
        py_threads: int,
        py_threads_idle_timeout: int,
        backpressure: int,
        http_mode: str,
        http1_opts: HTTP1Settings | None,
        http2_opts: HTTP2Settings | None,
        static_files: tuple[str, str, str | None, str | None] | None,
        ssl_enabled: bool,
        ssl_cert: str | None,
        ssl_key: str | None,
        ssl_key_password: str | None,
        ssl_protocol_min: str,
        ssl_ca: str | None,
        ssl_crl: list[str],
        ssl_client_verify: bool,
    ) -> WSGIWorker: ...

class RSGIWorker:
    def __new__(
        cls,
        worker_id: int,
        sock: Any,
        threads: int,
        blocking_threads: int,
        py_threads: int,
        py_threads_idle_timeout: int,
        backpressure: int,
        http_mode: str,
        http1_opts: HTTP1Settings | None,
        http2_opts: HTTP2Settings | None,
        websockets_enabled: bool,
        static_files: tuple[str, str, str | None, str | None] | None,
        ssl_enabled: bool,
        ssl_cert: str | None,
        ssl_key: str | None,
        ssl_key_password: str | None,
        ssl_protocol_min: str,
        ssl_ca: str | None,
        ssl_crl: list[str],
        ssl_client_verify: bool,
    ) -> RSGIWorker: ...

class SocketHolder:
    def get_fd(self) -> Any: ...
    def is_uds(self) -> bool: ...

class ListenerSpec:
    def __new__(cls, host: str, port: int, backlog: int) -> ListenerSpec: ...
    def build(self) -> SocketHolder: ...

if sys.platform != 'win32':
    class UnixListenerSpec:
        def __new__(cls, bind: str, backlog: int, permissions: int | None) -> UnixListenerSpec: ...
        def build(self) -> SocketHolder: ...
        def is_uds(self) -> bool: ...

class CallbackScheduler:
    _loop: Any
    _ctx: Any

    def _run(self, coro: Any) -> None: ...

class ProcInfoCollector:
    def __init__(self) -> None: ...
    def memory(self, pids: list[int] | None = None) -> int: ...

class IPCReceiverHandle:
    def __init__(self, id: int, fd: int): ...
    def run(self): ...

class IPCSenderHandle:
    def __init__(self, fd: int): ...

class MetricsAggregator:
    def __init__(self, size: int): ...
    def incr_spawn(self, val: int): ...
    def incr_respawn_err(self, val: int): ...
    def incr_respawn_ttl(self, val: int): ...
    def incr_respawn_rss(self, val: int): ...

class MetricsExporter:
    def __init__(self, aggregator: MetricsAggregator): ...
    def run(self, sock: SocketHolder, sig: WorkerSignal): ...



import sys
import threading
from collections.abc import Callable
from functools import wraps
from typing import Any

from .._futures import _future_watcher_wrapper, _new_cbscheduler
from .._granian import ASGIWorker, RSGIWorker, WorkerSignal, WorkerSignalSync, WSGIWorker
from .._loops import loops
from .._types import SSLCtx
from ..asgi import LifespanProtocol, _callback_wrapper as _asgi_call_wrap
from ..errors import ConfigurationError, FatalError
from ..rsgi import _callback_wrapper as _rsgi_call_wrap, _callbacks_from_target as _rsgi_cbs_from_target
from ..wsgi import _callback_wrapper as _wsgi_call_wrap
from .common import (
    WORKERS_METHODS,
    AbstractServer,
    AbstractWorker,
    HTTP1Settings,
    HTTP2Settings,
    HTTPModes,
    Interfaces,
    RuntimeModes,
    TaskImpl,
    logger,
)


class WorkerThread(AbstractWorker):
    _idl = 'TID'

    def __init__(self, parent, idx, target, args, sig):
        self._sig = sig
        super().__init__(parent, idx, target, args)

    @staticmethod
    def wrap_target(target):
        @wraps(target)
        def wrapped(worker_id, sig, callback, sock, loop_impl, *args, **kwargs):
            loop = loops.get(loop_impl)
            return target(worker_id, sig, callback, sock, loop, *args, **kwargs)

        return wrapped

    def _spawn(self, target, args):
        self.inner = threading.Thread(name='granian-worker', target=target, args=args)
        self._alive = True

    def _id(self):
        return self.inner.native_id

    def _watcher(self):
        self.inner.join()
        self._alive = False
        if not self.interrupt_by_parent:
            logger.error(f'Unexpected exit from worker-{self.idx + 1}')
            self.parent.interrupt_children.append(self.idx)
            self.parent.main_loop_interrupt.set()

    def terminate(self):
        self._alive = False
        self.interrupt_by_parent = True
        self._sig.set()

    def is_alive(self):
        if not self._alive:
            return False
        return self.inner.is_alive()


class MTServer(AbstractServer[WorkerThread]):
    @staticmethod
    @WorkerThread.wrap_target
    def _spawn_asgi_worker(
        worker_id: int,
        shutdown_event: Any,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
        metrics: Any,
    ):
        wcallback = _future_watcher_wrapper(_asgi_call_wrap(callback, scope_opts, {}, log_access_fmt))

        worker = ASGIWorker(
            worker_id,
            sock,
            None,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            metrics,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)

    @staticmethod
    @WorkerThread.wrap_target
    def _spawn_asgi_lifespan_worker(
        worker_id: int,
        shutdown_event: Any,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
        metrics: Any,
    ):
        lifespan_handler = LifespanProtocol(callback)
        wcallback = _future_watcher_wrapper(
            _asgi_call_wrap(callback, scope_opts, lifespan_handler.state, log_access_fmt)
        )

        loop.run_until_complete(lifespan_handler.startup())
        if lifespan_handler.interrupt:
            logger.error('ASGI lifespan startup failed', exc_info=lifespan_handler.exc)
            sys.exit(1)

        worker = ASGIWorker(
            worker_id,
            sock,
            None,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            metrics,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)
        loop.run_until_complete(lifespan_handler.shutdown())

    @staticmethod
    @WorkerThread.wrap_target
    def _spawn_rsgi_worker(
        worker_id: int,
        shutdown_event: Any,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
        metrics: Any,
    ):
        callback, callback_init, callback_del = _rsgi_cbs_from_target(callback)
        wcallback = _future_watcher_wrapper(_rsgi_call_wrap(callback, log_access_fmt))
        callback_init(loop)

        worker = RSGIWorker(
            worker_id,
            sock,
            None,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            metrics,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)
        callback_del(loop)

    @staticmethod
    @WorkerThread.wrap_target
    def _spawn_wsgi_worker(
        worker_id: int,
        shutdown_event: Any,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
        metrics: Any,
    ):
        wcallback = _wsgi_call_wrap(callback, scope_opts, log_access_fmt)

        worker = WSGIWorker(
            worker_id,
            sock,
            None,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            static_path,
            *ssl_ctx,
            metrics,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)

    def _spawn_worker(self, idx, target, callback_loader) -> WorkerThread:
        sig = WorkerSignalSync(threading.Event()) if self.interface == Interfaces.WSGI else WorkerSignal()

        return WorkerThread(
            parent=self,
            idx=idx,
            target=target,
            args=(
                idx + 1,
                sig,
                callback_loader,
                self._shd,
                self.loop,
                self.runtime_mode,
                self.runtime_threads,
                self.runtime_blocking_threads,
                self.blocking_threads,
                self.blocking_threads_idle_timeout,
                self.backpressure,
                self.task_impl,
                self.http,
                self.http1_settings,
                self.http2_settings,
                self.websockets,
                self.static_path,
                self.log_access_format if self.log_access else None,
                self.ssl_ctx,
                {'url_path_prefix': self.url_path_prefix},
                (self.metrics_scrape_interval if self.metrics_enabled else None, self._metrics),
            ),
            sig=sig,
        )

    def _check_gil(self):
        try:
            assert sys._is_gil_enabled() is False
        except Exception:
            logger.error('Cannot run a free-threaded Granian build with GIL enabled')
            raise FatalError('gil')

    def _serve(self, spawn_target, target_loader):
        target = target_loader()
        self._check_gil()
        self.startup(spawn_target, target)
        self._serve_loop(spawn_target, target)
        self.shutdown()

    def _serve_with_reloader(self, spawn_target, target_loader):
        raise NotImplementedError

    def serve(
        self,
        spawn_target: Callable[..., None] | None = None,
        target_loader: Callable[..., Callable[..., Any]] | None = None,
        wrap_loader: bool = True,
    ):
        logger.warning('free-threaded Python support is experimental!')

        if self.reload_on_changes:
            logger.error('The changes reloader is not supported on the free-threaded build')
            raise ConfigurationError('reload')

        if self.workers_rss:
            logger.error('The resource monitor is not supported on the free-threaded build')
            raise ConfigurationError('workers_max_rss')

        super().serve(spawn_target, target_loader, wrap_loader)



import multiprocessing
import os
import socket
import sys
from collections.abc import Callable
from functools import wraps
from typing import Any

from .._futures import _future_watcher_wrapper, _new_cbscheduler
from .._granian import (
    ASGIWorker,
    IPCReceiverHandle,
    IPCSenderHandle,
    ProcInfoCollector,
    RSGIWorker,
    SocketHolder,
    WorkerSignal,
    WSGIWorker,
)
from .._internal import load_env
from .._types import SSLCtx
from ..asgi import LifespanProtocol, _callback_wrapper as _asgi_call_wrap
from ..rsgi import _callback_wrapper as _rsgi_call_wrap, _callbacks_from_target as _rsgi_cbs_from_target
from ..wsgi import _callback_wrapper as _wsgi_call_wrap
from .common import (
    WORKERS_METHODS,
    AbstractServer,
    AbstractWorker,
    HTTP1Settings,
    HTTP2Settings,
    HTTPModes,
    Interfaces,
    RuntimeModes,
    TaskImpl,
    configure_logging,
    logger,
    setproctitle,
)


multiprocessing.allow_connection_pickling()


class WorkerProcess(AbstractWorker):
    _idl = 'PID'

    def __init__(self, parent, idx, target, args):
        # NOTE: Python 3.14 defaults mp spawn method to 'forkserver' on Linux,
        #       which doesn't really play well with shared sockets.
        self._spawn_method = multiprocessing.get_start_method()
        if self._spawn_method not in {'fork', 'spawn'}:
            self._spawn_method = 'spawn'
        super().__init__(parent, idx, target, args)

    @staticmethod
    def wrap_target(target):
        @wraps(target)
        def wrapped(
            worker_id,
            process_name,
            callback_loader,
            sock,
            ipc,
            loop_impl,
            log_enabled,
            log_level,
            log_config,
            env_files,
            *args,
            **kwargs,
        ):
            from granian._loops import loops

            if process_name:
                setproctitle.setproctitle(f'{process_name} worker-{worker_id}')

            configure_logging(log_level, log_config, log_enabled)
            load_env(env_files)

            _ipc_handle = None
            sock, _sso = sock
            if sys.platform == 'win32':
                sock = SocketHolder(_sso.fileno())
            elif ipc:
                _ipc_fd = os.dup(ipc.fileno())
                os.set_blocking(_ipc_fd, False)
                _ipc_handle = IPCSenderHandle(_ipc_fd)

            loop = loops.get(loop_impl)
            callback = callback_loader()
            return target(worker_id, callback, sock, _ipc_handle, loop, *args, **kwargs)

        return wrapped

    def _spawn(self, target, args):
        self.inner = multiprocessing.get_context(self._spawn_method).Process(
            name='granian-worker', target=target, args=args
        )

    def _id(self):
        return self.inner.pid

    def terminate(self):
        self.interrupt_by_parent = True
        self.inner.terminate()

    def kill(self):
        self.interrupt_by_parent = True
        self.inner.kill()


class MPServer(AbstractServer[WorkerProcess]):
    @staticmethod
    @WorkerProcess.wrap_target
    def _spawn_asgi_worker(
        worker_id: int,
        callback: Any,
        sock: Any,
        ipc: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
        metrics: Any,
    ):
        from granian._signals import set_loop_signals

        wcallback = _future_watcher_wrapper(_asgi_call_wrap(callback, scope_opts, {}, log_access_fmt))
        shutdown_event = set_loop_signals(loop)

        worker = ASGIWorker(
            worker_id,
            sock,
            ipc,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            metrics,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)

    @staticmethod
    @WorkerProcess.wrap_target
    def _spawn_asgi_lifespan_worker(
        worker_id: int,
        callback: Any,
        sock: Any,
        ipc: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
        metrics: Any,
    ):
        from granian._signals import set_loop_signals

        lifespan_handler = LifespanProtocol(callback)
        wcallback = _future_watcher_wrapper(
            _asgi_call_wrap(callback, scope_opts, lifespan_handler.state, log_access_fmt)
        )
        shutdown_event = set_loop_signals(loop)

        loop.run_until_complete(lifespan_handler.startup())
        if lifespan_handler.interrupt:
            logger.error('ASGI lifespan startup failed', exc_info=lifespan_handler.exc)
            sys.exit(1)

        worker = ASGIWorker(
            worker_id,
            sock,
            ipc,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            metrics,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)
        loop.run_until_complete(lifespan_handler.shutdown())

    @staticmethod
    @WorkerProcess.wrap_target
    def _spawn_rsgi_worker(
        worker_id: int,
        callback: Any,
        sock: Any,
        ipc: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
        metrics: Any,
    ):
        from granian._signals import set_loop_signals

        callback, callback_init, callback_del = _rsgi_cbs_from_target(callback)
        wcallback = _future_watcher_wrapper(_rsgi_call_wrap(callback, log_access_fmt))
        shutdown_event = set_loop_signals(loop)
        callback_init(loop)

        worker = RSGIWorker(
            worker_id,
            sock,
            ipc,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            metrics,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)
        callback_del(loop)

    @staticmethod
    @WorkerProcess.wrap_target
    def _spawn_wsgi_worker(
        worker_id: int,
        callback: Any,
        sock: Any,
        ipc: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
        metrics: Any,
    ):
        from granian._signals import set_sync_signals

        wcallback = _wsgi_call_wrap(callback, scope_opts, log_access_fmt)
        shutdown_event = set_sync_signals()

        worker = WSGIWorker(
            worker_id,
            sock,
            ipc,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            static_path,
            *ssl_ctx,
            metrics,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)

    def _init_shared_socket(self):
        super()._init_shared_socket()
        sock = socket.socket(fileno=self._sfd)
        sock.set_inheritable(True)
        self._sso = sock

    def _write_pidfile(self):
        super()._write_pidfile()
        self._rss_collector = ProcInfoCollector()

    def _unlink_pidfile(self):
        self._sso.detach()
        super()._unlink_pidfile()

    def _start_ipc(self):
        self._ipc = {}
        self._ipc_sig = WorkerSignal()

        # NOTE: for "reasons", on Windows the call to `os.set_blocking` fails with an OS err 9.
        #       I'm not sure what the problem is, thus on Windows we just disable
        #       IPC – and, consequentially, metrics – entirely; or, at least,
        #       until until someone finds out a way to make that call work.
        #       And, to quote John Malkovich: "fuck Microsoft!"
        #       (https://www.youtube.com/watch?v=2zpCOYkdvTQ)
        if sys.platform == 'win32':
            for idx in range(self.workers):
                self._ipc[idx] = (None, None)
            return

        for idx in range(self.workers):
            rx, tx = multiprocessing.Pipe(False)
            rxd = os.dup(rx.fileno())
            # WARN: on Windows, `os.set_blocking` is available only on Py >= 3.12.
            #       Doesn't really matter, given the call fails when available U.U
            os.set_blocking(rxd, False)
            self._ipc[idx] = (IPCReceiverHandle(idx, rxd), tx, rx)

        # NOTE: given we use IPC only for metrics right now, let's run the receivers
        #       only if metrics collection is actually enabled.
        if self.metrics_enabled:
            for pipe in self._ipc.values():
                pipe[0].run(self._ipc_sig, self._metrics)

    def _stop_ipc(self):
        self._ipc_sig.set()

    def _handle_rss_signal(self, spawn_target, target_loader):
        wpids = {wrk._id(): wrk for wrk in self.wrks}
        try:
            rss_data = self._rss_collector.memory(list(wpids.keys()))
        except Exception:
            logger.warning('Unable to collect resource usage for workers')
            return
        logger.debug(f'Collected resource usages for workers: {rss_data}')
        cycle_samples = {}
        to_restart = []
        for wpid, wmem in rss_data.items():
            if wmem >= self.workers_rss:
                samples = self._rss_wrk_samples.get(wpid, 0) + 1
                if samples >= self.rss_samples:
                    wrk = wpids[wpid]
                    logger.info(f'worker-{wrk.idx + 1} RSS over threshold, gracefully respawning..')
                    to_restart.append(wrk.idx)
                else:
                    cycle_samples[wpid] = samples
            else:
                cycle_samples[wpid] = 0
        self._rss_wrk_samples.clear()
        self._rss_wrk_samples.update(cycle_samples)
        if to_restart:
            self._respawn_workers(to_restart, spawn_target, target_loader, delay=self.respawn_interval)
            self._metrics.incr_respawn_rss(len(to_restart))

    def _spawn_worker(self, idx, target, callback_loader) -> WorkerProcess:
        return WorkerProcess(
            parent=self,
            idx=idx,
            target=target,
            args=(
                idx + 1,
                self.process_name,
                callback_loader,
                (self._shd, self._sso),
                # NOTE: given we use IPC only for metrics right now, let's share the pipe
                #       only if metrics collection is actually enabled.
                self._ipc[idx][1] if self.metrics_enabled else None,
                self.loop,
                self.log_enabled,
                self.log_level,
                self.log_config,
                self.env_files,
                self.runtime_mode,
                self.runtime_threads,
                self.runtime_blocking_threads,
                self.blocking_threads,
                self.blocking_threads_idle_timeout,
                self.backpressure,
                self.task_impl,
                self.http,
                self.http1_settings,
                self.http2_settings,
                self.websockets,
                self.static_path,
                self.log_access_format if self.log_access else None,
                self.ssl_ctx,
                {'url_path_prefix': self.url_path_prefix},
                (self.metrics_scrape_interval if self.metrics_enabled else None, None),
            ),
        )

    def serve(
        self,
        spawn_target: Callable[..., None] | None = None,
        target_loader: Callable[..., Callable[..., Any]] | None = None,
        wrap_loader: bool = True,
    ):
        if self.interface == Interfaces.WSGI:
            if self.blocking_threads > (multiprocessing.cpu_count() * 2 + 1):
                logger.warning(
                    f'Configuration allows spawning up to {self.blocking_threads} Python threads, '
                    'which seems quite high compared to the number of CPU cores available. '
                    'Consider reviewing your configuration and using `backpressure` to limit '
                    'the concurrency on the Python interpreter. '
                    'If this configuration is intentional, you can safely ignore this message.'
                )

        if self.metrics_enabled and sys.platform == 'win32':
            self.metrics_enabled = False
            logger.warn('Metrics are not available in Windows, ignoring.')

        super().serve(spawn_target, target_loader, wrap_loader)



import json
import pathlib
import socket
import ssl

import httpx
import pytest
import websockets


@pytest.mark.asyncio
@pytest.mark.parametrize('server_tls', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
@pytest.mark.parametrize(
    'tls_scope', [('tls1.2', ssl.TLSVersion.TLSv1_2, 'TLSv1.2'), ('tls1.3', ssl.TLSVersion.TLSv1_3, 'TLSv1.3')]
)
async def test_tls_protocol(server_tls, runtime_mode, tls_scope):
    tls_input, tls_max_proto, tls_expected = tls_scope

    async with server_tls(runtime_mode, ws=False, tls_proto=tls_input) as port:
        context = ssl.create_default_context()
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        context.maximum_version = tls_max_proto

        with socket.create_connection(('localhost', port)) as sock:
            with context.wrap_socket(sock, server_hostname='localhost') as ssock:
                tls_version = ssock.version()

        assert tls_version == tls_expected


@pytest.mark.asyncio
@pytest.mark.parametrize('server_tls', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_http_scope(server_tls, runtime_mode):
    async with server_tls(runtime_mode, ws=False) as port:
        res = httpx.get(f'https://localhost:{port}/info?test=true', verify=False)

    assert res.status_code == 200
    data = res.json()
    assert data['scheme'] == 'https'


@pytest.mark.asyncio
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_asgi_ws_scope(asgi_server, runtime_mode):
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    localhost_pem = pathlib.Path.cwd() / 'tests' / 'fixtures' / 'tls' / 'cert.pem'
    ssl_context.load_verify_locations(str(localhost_pem))

    async with asgi_server(runtime_mode, tls=True) as port:
        async with websockets.connect(f'wss://localhost:{port}/ws_info?test=true', ssl=ssl_context) as ws:
            res = await ws.recv()

    data = json.loads(res)
    assert data['scheme'] == 'wss'


@pytest.mark.asyncio
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_rsgi_ws_scope(rsgi_server, runtime_mode):
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    localhost_pem = pathlib.Path.cwd() / 'tests' / 'fixtures' / 'tls' / 'cert.pem'
    ssl_context.load_verify_locations(str(localhost_pem))

    async with rsgi_server(runtime_mode, tls=True) as port:
        async with websockets.connect(f'wss://localhost:{port}/ws_info?test=true', ssl=ssl_context) as ws:
            res = await ws.recv()

    data = json.loads(res)
    assert data['scheme'] == 'https'


@pytest.mark.asyncio
async def test_tls_encrypted_key(rsgi_server):
    async with rsgi_server('st', ws=False, tls='priv') as port:
        res = httpx.get(f'https://localhost:{port}/info?test=true', verify=False)

    assert res.status_code == 200
    data = res.json()
    assert data['scheme'] == 'https'



import asyncio
import multiprocessing
import sys
import time
from collections.abc import Callable, Sequence
from functools import wraps
from pathlib import Path
from typing import Any

from .._futures import _future_watcher_wrapper, _new_cbscheduler
from .._granian import ASGIWorker, RSGIWorker, WorkerSignal
from .._imports import dotenv
from .._internal import load_env
from .._types import SSLCtx
from ..asgi import LifespanProtocol, _callback_wrapper as _asgi_call_wrap
from ..errors import ConfigurationError, FatalError
from ..rsgi import _callback_wrapper as _rsgi_call_wrap, _callbacks_from_target as _rsgi_cbs_from_target
from .common import (
    _PY_312,
    _PYV,
    AbstractServer,
    AbstractWorker,
    HTTP1Settings,
    HTTP2Settings,
    HTTPModes,
    Interfaces,
    LogLevels,
    SSLProtocols,
    TaskImpl,
    logger,
)


class AsyncWorker(AbstractWorker):
    def __init__(self, parent, idx, target, args, sig):
        self._sig = sig
        self._loop = asyncio.get_event_loop()
        self._task = None
        self._wtask = None
        super().__init__(parent, idx, target, args)

    @staticmethod
    def wrap_target(target):
        @wraps(target)
        def wrapped(worker_id, sig, callback, sock, *args, **kwargs):
            loop = asyncio.get_event_loop()
            return target(worker_id, sig, callback, sock, loop, *args, **kwargs)

        return wrapped

    def _spawn(self, target, args):
        self._task = self._loop.create_task(target(*args))
        self._alive = True

    def _id(self):
        return id(self._task)

    async def _watcher(self):
        try:
            await self._task
        except BaseException:
            pass
        if not self.interrupt_by_parent:
            logger.error(f'Unexpected exit from worker-{self.idx + 1}')
            self.parent.interrupt_children.append(self.idx)
            self.parent.main_loop_interrupt.set()

    def _watch(self):
        self._wtask = self._loop.create_task(self._watcher())

    def start(self):
        logger.info(f'Spawning worker-{self.idx + 1} with {self._idl}: {self._id()}')
        self._watch()

    def is_alive(self):
        if not self._alive:
            return False
        return not self._task.done()

    def terminate(self):
        self._alive = False
        self.interrupt_by_parent = True
        self._sig.set()

    def kill(self):
        self._alive = False
        self.interrupt_by_parent = True
        self._task.cancel()

    def join(self, timeout=None):
        return asyncio.wait_for(self._task, timeout=timeout)


class Server(AbstractServer[AsyncWorker]):
    def __init__(
        self,
        target: Any,
        address: str = '127.0.0.1',
        port: int = 8000,
        uds: Path | None = None,
        interface: Interfaces = Interfaces.RSGI,
        blocking_threads: int | None = None,
        blocking_threads_idle_timeout: int = 30,
        runtime_threads: int = 1,
        runtime_blocking_threads: int | None = None,
        task_impl: TaskImpl = TaskImpl.asyncio,
        http: HTTPModes = HTTPModes.auto,
        websockets: bool = True,
        backlog: int = 128,
        backpressure: int | None = None,
        http1_settings: HTTP1Settings | None = None,
        http2_settings: HTTP2Settings | None = None,
        log_enabled: bool = True,
        log_level: LogLevels = LogLevels.info,
        log_dictconfig: dict[str, Any] | None = None,
        log_access: bool = False,
        log_access_format: str | None = None,
        ssl_cert: Path | None = None,
        ssl_key: Path | None = None,
        ssl_key_password: str | None = None,
        ssl_protocol_min: SSLProtocols = SSLProtocols.tls13,
        ssl_ca: Path | None = None,
        ssl_crl: list[Path] | None = None,
        ssl_client_verify: bool = False,
        url_path_prefix: str | None = None,
        factory: bool = False,
        static_path_route: Sequence[str] | None = None,
        static_path_mount: Sequence[Path] | None = None,
        static_path_dir_to_file: str | None = None,
        static_path_expires: int = 86400,
    ):
        super().__init__(
            target=target,
            address=address,
            port=port,
            uds=uds,
            interface=interface,
            blocking_threads=blocking_threads,
            blocking_threads_idle_timeout=blocking_threads_idle_timeout,
            runtime_threads=runtime_threads,
            runtime_blocking_threads=runtime_blocking_threads,
            task_impl=task_impl,
            http=http,
            websockets=websockets,
            backlog=backlog,
            backpressure=backpressure,
            http1_settings=http1_settings,
            http2_settings=http2_settings,
            log_enabled=log_enabled,
            log_level=log_level,
            log_dictconfig=log_dictconfig,
            log_access=log_access,
            log_access_format=log_access_format,
            ssl_cert=ssl_cert,
            ssl_key=ssl_key,
            ssl_key_password=ssl_key_password,
            ssl_protocol_min=ssl_protocol_min,
            ssl_ca=ssl_ca,
            ssl_crl=ssl_crl,
            ssl_client_verify=ssl_client_verify,
            url_path_prefix=url_path_prefix,
            factory=factory,
            static_path_route=static_path_route,
            static_path_mount=static_path_mount,
            static_path_dir_to_file=static_path_dir_to_file,
            static_path_expires=static_path_expires,
        )
        self.main_loop_interrupt = asyncio.Event()

    def _spawn_worker(self, idx, target, callback_loader) -> AsyncWorker:
        sig = WorkerSignal()

        return AsyncWorker(
            parent=self,
            idx=idx,
            target=target,
            args=(
                idx + 1,
                sig,
                callback_loader,
                self._shd,
                self.runtime_threads,
                self.runtime_blocking_threads,
                self.blocking_threads,
                self.blocking_threads_idle_timeout,
                self.backpressure,
                self.task_impl,
                self.http,
                self.http1_settings,
                self.http2_settings,
                self.websockets,
                self.static_path,
                self.log_access_format if self.log_access else None,
                self.ssl_ctx,
                {'url_path_prefix': self.url_path_prefix},
            ),
            sig=sig,
        )

    @staticmethod
    @AsyncWorker.wrap_target
    async def _spawn_asgi_worker(
        worker_id: int,
        shutdown_event: Any,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
    ):
        wcallback = _future_watcher_wrapper(_asgi_call_wrap(callback, scope_opts, {}, log_access_fmt))

        worker = ASGIWorker(
            worker_id,
            sock,
            None,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            (None, None),
        )
        serve = worker.serve_async_uds if sock.is_uds() else worker.serve_async
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        await serve(scheduler, loop, shutdown_event)

    @staticmethod
    @AsyncWorker.wrap_target
    async def _spawn_asgi_lifespan_worker(
        worker_id: int,
        shutdown_event: Any,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
    ):
        lifespan_handler = LifespanProtocol(callback)
        wcallback = _future_watcher_wrapper(
            _asgi_call_wrap(callback, scope_opts, lifespan_handler.state, log_access_fmt)
        )

        await lifespan_handler.startup()
        if lifespan_handler.interrupt:
            logger.error('ASGI lifespan startup failed', exc_info=lifespan_handler.exc)
            raise FatalError('ASGI lifespan startup')

        worker = ASGIWorker(
            worker_id,
            sock,
            None,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            (None, None),
        )
        serve = worker.serve_async_uds if sock.is_uds() else worker.serve_async
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        await serve(scheduler, loop, shutdown_event)
        await lifespan_handler.shutdown()

    @staticmethod
    @AsyncWorker.wrap_target
    async def _spawn_rsgi_worker(
        worker_id: int,
        shutdown_event: Any,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_threads: int,
        runtime_blocking_threads: int | None,
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: HTTP1Settings | None,
        http2_settings: HTTP2Settings | None,
        websockets: bool,
        static_path: tuple[str, str, str | None, str | None] | None,
        log_access_fmt: str | None,
        ssl_ctx: SSLCtx,
        scope_opts: dict[str, Any],
    ):
        callback, callback_init, callback_del = _rsgi_cbs_from_target(callback)
        wcallback = _future_watcher_wrapper(_rsgi_call_wrap(callback, log_access_fmt))
        callback_init(loop)

        worker = RSGIWorker(
            worker_id,
            sock,
            None,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
            (None, None),
        )
        serve = worker.serve_async_uds if sock.is_uds() else worker.serve_async
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        await serve(scheduler, loop, shutdown_event)
        callback_del(loop)

    async def _respawn_workers(self, workers, spawn_target, target_loader, delay: float = 0):
        for idx in workers:
            self.respawned_wrks[idx] = time.monotonic()
            logger.info(f'Respawning worker-{idx + 1}')
            old_wrk = self.wrks.pop(idx)
            wrk = self._spawn_worker(idx=idx, target=spawn_target, callback_loader=target_loader)
            wrk.start()
            self.wrks.insert(idx, wrk)
            await asyncio.sleep(delay)
            logger.info(f'Stopping old worker-{idx + 1}')
            old_wrk.terminate()
            await old_wrk.join(self.workers_kill_timeout)
            if self.workers_kill_timeout:
                # the worker might still be reported alive after `join`, let's context switch
                if old_wrk.is_alive():
                    await asyncio.sleep(0.001)
                if old_wrk.is_alive():
                    logger.warning(f'Killing old worker-{idx + 1} after it refused to gracefully stop')
                    old_wrk.kill()
                    await old_wrk.join()

    async def _stop_workers(self):
        for wrk in self.wrks:
            wrk.terminate()

        for wrk in self.wrks:
            await wrk.join(self.workers_kill_timeout)
            if self.workers_kill_timeout:
                if wrk.is_alive():
                    logger.warning(f'Killing worker-{wrk.idx} after it refused to gracefully stop')
                    wrk.kill()

        self.wrks.clear()

    def startup(self, spawn_target, target_loader):
        logger.info('Starting granian (embedded)')
        self._init_shared_socket()
        proto = 'https' if self.ssl_ctx[0] else 'http'
        logger.info(f'Listening at: {proto}://{self._bind_addr_fmt}')

        load_env(self.env_files)
        self._call_hooks(self.hooks_startup)
        self._spawn_workers(spawn_target, target_loader)

    async def _serve_loop(self, spawn_target, target_loader):
        while True:
            await self.main_loop_interrupt.wait()
            if self.interrupt_signal:
                break

            if self.interrupt_children:
                break

            if self.reload_signal:
                await self._reload(spawn_target, target_loader)

    async def shutdown(self, exit_code=0):
        logger.info('Shutting down granian')
        await self._stop_workers()
        self._call_hooks(self.hooks_shutdown)
        if self.bind_uds and self.bind_uds.exists():
            self.bind_uds.unlink()

    async def _serve(self, spawn_target, target_loader):
        target = target_loader()
        self.startup(spawn_target, target)
        await self._serve_loop(spawn_target, target)
        await self.shutdown()

    async def serve(self, spawn_target: Callable[..., None] | None = None):
        def target_loader(*args, **kwargs):
            if self.factory:
                return self.target()
            return self.target

        default_spawners = {
            Interfaces.ASGI: self._spawn_asgi_lifespan_worker,
            Interfaces.ASGINL: self._spawn_asgi_worker,
            Interfaces.RSGI: self._spawn_rsgi_worker,
        }

        logger.warning('Embedded server is experimental!')

        if self.interface == Interfaces.WSGI:
            logger.error('WSGI is not supported in embedded mode')
            raise ConfigurationError('interface')

        if self.reload_on_changes:
            logger.error('The changes reloader is not supported in embedded mode')
            raise ConfigurationError('reload')

        if self.workers_rss:
            logger.error('The resource monitor is not supported in embedded mode')
            raise ConfigurationError('workers_max_rss')

        if self.metrics_enabled:
            logger.error('Metrics are not available in embedded mode')
            raise ConfigurationError('metrics_enabled')

        if not spawn_target:
            spawn_target = default_spawners[self.interface]

        if self.bind_uds and sys.platform == 'win32':
            logger.error('Unix Domain sockets are not available on Windows')
            raise ConfigurationError('uds')

        if self.blocking_threads > 1:
            logger.error('Blocking threads > 1 is not supported on ASGI and RSGI')
            raise ConfigurationError('blocking_threads')

        if self.websockets:
            if self.http == HTTPModes.http2:
                logger.info('Websockets are not supported on HTTP/2 only, ignoring')

        if self.env_files and dotenv is None:
            logger.error('Environment file(s) usage requires the granian[dotenv] extra')
            raise ConfigurationError('env_files')

        if self.blocking_threads_idle_timeout < 5 or self.blocking_threads_idle_timeout > 600:
            logger.error('Blocking threads idle timeout must be between 10 and 600 seconds')
            raise ConfigurationError('blocking_threads_idle_timeout')

        cpus = multiprocessing.cpu_count()
        if self.runtime_threads > cpus:
            logger.warning(
                'Configured number of Rust threads appears to be too high given the amount of CPU cores available. '
                'Mind that Rust threads are not involved in Python code execution, and they almost never be the '
                'limiting factor in scaling. Consider configuring the amount of blocking threads instead'
            )

        if self.task_impl == TaskImpl.rust:
            if _PYV >= _PY_312:
                self.task_impl = TaskImpl.asyncio
                logger.warning('Rust task implementation is not available on Python >= 3.12, falling back to asyncio')
            else:
                logger.warning('Rust task implementation is experimental!')

        await self._serve(spawn_target, target_loader)

    def stop(self):
        self.signal_handler_interrupt()

    def reload(self):
        self.signal_handler_reload()



import os
import sys
import time
from collections.abc import Callable
from functools import wraps
from typing import Any

from .log import log_request_builder


class Response:
    __slots__ = ['status', 'headers']

    def __init__(self):
        self.status = 200
        self.headers = []

    def __call__(self, status: str, headers: list[tuple[str, str]], exc_info: Any = None):
        self.status = int(status.split(' ', 1)[0])
        self.headers = headers


class ResponseIterWrap:
    __slots__ = ['inner', '__next__']

    def __init__(self, inner):
        self.inner = inner
        self.__next__ = iter(inner).__next__

    def close(self):
        self.inner.close()


def _callback_wrapper(callback: Callable[..., Any], scope_opts: dict[str, Any], access_log_fmt=None):
    basic_env: dict[str, Any] = dict(os.environ)
    basic_env.update(
        {
            'GATEWAY_INTERFACE': 'CGI/1.1',
            'SCRIPT_NAME': scope_opts.get('url_path_prefix') or '',
            'SERVER_SOFTWARE': 'Granian',
            'wsgi.errors': sys.stderr,
            'wsgi.multiprocess': False,
            'wsgi.multithread': True,
            'wsgi.run_once': False,
            'wsgi.version': (1, 0),
        }
    )

    def _runner(proto, scope):
        resp = Response()
        scope.update(basic_env)
        if scope['SCRIPT_NAME']:
            scope['PATH_INFO'] = scope['PATH_INFO'][len(scope['SCRIPT_NAME']) :] or '/'

        rv = callback(scope, resp)

        if isinstance(rv, list):
            proto.response_bytes(resp.status, resp.headers, b''.join(rv))
        else:
            proto.response_iter(resp.status, resp.headers, ResponseIterWrap(rv))

        return resp.status

    def _logger(proto, scope):
        rt, mt = time.time(), time.perf_counter()
        try:
            status = _runner(proto, scope)
            access_log(rt, mt, scope, status)
        except BaseException:
            access_log(rt, mt, scope, 500)
            raise
        return status

    access_log = _build_access_logger(access_log_fmt)
    wrapper = _logger if access_log_fmt else _runner
    wraps(callback)(wrapper)
    return wrapper


def _build_access_logger(fmt):
    logger = log_request_builder(fmt)

    def access_log(rt, mt, scope, resp_code):
        logger(
            rt,
            mt,
            {
                'addr_remote': scope['REMOTE_ADDR'].rsplit(':', 1)[0],
                'protocol': scope['SERVER_PROTOCOL'],
                'path': scope['PATH_INFO'],
                'qs': scope['QUERY_STRING'],
                'method': scope['REQUEST_METHOD'],
                'scheme': scope['wsgi.url_scheme'],
            },
            resp_code,
        )

    return access_log



import httpx
import pytest


@pytest.mark.asyncio
@pytest.mark.parametrize('server_static_files', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
@pytest.mark.parametrize('file_name', ['media.png', 'こんにちは.png'])
async def test_static_files(server_static_files, runtime_mode, file_name):
    async with server_static_files(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/static/{file_name}')

    assert res.status_code == 200
    assert res.headers.get('content-type') == 'image/png'
    assert res.headers.get('cache-control')


@pytest.mark.asyncio
@pytest.mark.parametrize('server_static_files', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_static_files_notfound(server_static_files, runtime_mode):
    async with server_static_files(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/static/missing.png')

    assert res.status_code == 404


@pytest.mark.asyncio
@pytest.mark.parametrize('server_static_files', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_static_files_dir_no_rewrite(server_static_files, runtime_mode):
    async with server_static_files(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/static/empty')

    assert res.status_code == 404


@pytest.mark.asyncio
@pytest.mark.parametrize('server_static_files', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_static_files_outsidemount(monkeypatch, server_static_files, runtime_mode):
    monkeypatch.setattr(httpx._urlparse, 'normalize_path', lambda v: v)

    async with server_static_files(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/static/../conftest.py')

    assert res.status_code == 404


@pytest.mark.asyncio
@pytest.mark.parametrize('server_static_files', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_static_files_approute(server_static_files, runtime_mode):
    async with server_static_files(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/info')

    assert res.status_code == 200


@pytest.mark.asyncio
@pytest.mark.parametrize('server_static_files', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_static_files_rewrite_index(server_static_files, runtime_mode):
    async with server_static_files(runtime_mode, ws=False, static_rewrite=True) as port:
        res = httpx.get(f'http://localhost:{port}/static/file_rewrite')

    assert res.status_code == 200


@pytest.mark.asyncio
@pytest.mark.parametrize('server_static_files', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_static_files_rewrite_notfound(server_static_files, runtime_mode):
    async with server_static_files(runtime_mode, ws=False, static_rewrite=True) as port:
        res = httpx.get(f'http://localhost:{port}/static/empty')

    assert res.status_code == 404


@pytest.mark.asyncio
@pytest.mark.parametrize('server_static_files', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_static_files_multi(server_static_files, runtime_mode):
    async with server_static_files(
        runtime_mode, ws=False, static_mount=[('/static', 'static'), ('/other', 'tls')]
    ) as port:
        res1 = httpx.get(f'http://localhost:{port}/static/media.png')
        res2 = httpx.get(f'http://localhost:{port}/other/cert.pem')

    assert res1.status_code == 200
    assert res2.status_code == 200
    assert res1.headers.get('content-type') == 'image/png'
    assert res2.headers.get('content-type') == 'application/x-x509-ca-cert'



import copy
import datetime
import logging
import logging.config
import time
from enum import Enum
from typing import Any


class LogLevels(str, Enum):
    critical = 'critical'
    error = 'error'
    warning = 'warning'
    warn = 'warn'
    info = 'info'
    debug = 'debug'
    notset = 'notset'


log_levels_map = {
    LogLevels.critical: logging.CRITICAL,
    LogLevels.error: logging.ERROR,
    LogLevels.warning: logging.WARNING,
    LogLevels.warn: logging.WARN,
    LogLevels.info: logging.INFO,
    LogLevels.debug: logging.DEBUG,
    LogLevels.notset: logging.NOTSET,
}

LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'generic': {
            '()': 'logging.Formatter',
            'fmt': '[%(levelname)s] %(message)s',
            'datefmt': '[%Y-%m-%d %H:%M:%S %z]',
        },
        'access': {
            '()': 'logging.Formatter',
            'fmt': '%(message)s',
            'datefmt': '[%Y-%m-%d %H:%M:%S %z]',
        },
    },
    'handlers': {
        'console': {'formatter': 'generic', 'class': 'logging.StreamHandler', 'stream': 'ext://sys.stdout'},
        'access': {'formatter': 'access', 'class': 'logging.StreamHandler', 'stream': 'ext://sys.stdout'},
    },
    'loggers': {
        '_granian': {'handlers': ['console'], 'level': 'INFO', 'propagate': False},
        'granian.access': {'handlers': ['access'], 'level': 'INFO', 'propagate': False},
    },
}

DEFAULT_ACCESSLOG_FMT = '[%(time)s] %(addr)s - "%(method)s %(path)s %(protocol)s" %(status)d %(dt_ms).3f'

# NOTE: to be consistent with the Rust module logger name
logger = logging.getLogger('_granian')
access_logger = logging.getLogger('granian.access')


def configure_logging(level: LogLevels, config: dict[str, Any] | None = None, enabled: bool = True):
    log_config = copy.deepcopy(LOGGING_CONFIG)

    if config:
        log_config.update(config)

    log_config['loggers'].setdefault('_granian', {})['level'] = log_levels_map[level]
    logging.config.dictConfig(log_config)

    if not enabled:
        logger.setLevel(logging.CRITICAL + 1)


def log_request_builder(fmt):
    now = datetime.datetime.now()
    local_now = now.astimezone()
    local_tz = local_now.tzinfo

    def log_request(rtime, mtime, req, res_code):
        dt = time.perf_counter() - mtime
        rdt = datetime.datetime.fromtimestamp(rtime, tz=local_tz)
        access_logger.info(
            fmt,
            {
                'addr': req['addr_remote'],
                'time': rdt.strftime('%Y-%m-%d %H:%M:%S %z'),
                'dt_ms': dt * 1000,
                'status': res_code,
                'path': req['path'],
                'query_string': req['qs'],
                'method': req['method'],
                'scheme': req['scheme'],
                'protocol': req['protocol'],
            },
        )

    return log_request



import signal
import sys
import threading

from ._granian import WorkerSignal, WorkerSignalSync


def _get_signals():
    rv = [signal.SIGINT, signal.SIGTERM]
    if sys.platform == 'win32':
        rv.append(signal.SIGBREAK)
    return rv


def set_main_signals(interrupt_handler, reload_handler=None):
    for sig in _get_signals():
        signal.signal(sig, interrupt_handler)

    if reload_handler is not None and sys.platform != 'win32':
        signal.signal(signal.SIGHUP, reload_handler)


def set_loop_signals(loop):
    signal_event = WorkerSignal()

    def signal_handler(signum, frame):
        signal_event.set()

    signals = _get_signals()
    try:
        for sigval in signals:
            loop.add_signal_handler(sigval, signal_handler, sigval, None)
    except NotImplementedError:
        for sigval in signals:
            signal.signal(sigval, signal_handler)

    return signal_event


def set_sync_signals():
    signal_event = WorkerSignalSync(threading.Event())

    def signal_handler(signum, frame):
        signal_event.set()

    set_main_signals(signal_handler)
    return signal_event



from __future__ import annotations

import errno
import multiprocessing
import os
import ssl
import sys
import threading
import time
from collections.abc import Callable, Sequence
from functools import partial
from pathlib import Path
from typing import Any, Generic, TypeVar

from .._compat import _PY_312, _PYV
from .._granian import MetricsAggregator, MetricsExporter, WorkerSignal
from .._imports import dotenv, setproctitle, watchfiles
from .._internal import build_env_loader, load_target
from .._signals import set_main_signals
from ..constants import HTTPModes, Interfaces, Loops, RuntimeModes, SSLProtocols, TaskImpl
from ..errors import ConfigurationError, PidFileError
from ..http import HTTP1Settings, HTTP2Settings
from ..log import DEFAULT_ACCESSLOG_FMT, LogLevels, configure_logging, logger
from ..net import SocketSpec, UnixSocketSpec


WT = TypeVar('WT')

WORKERS_METHODS = {
    RuntimeModes.mt: {False: 'serve_mtr', True: 'serve_mtr_uds'},
    RuntimeModes.st: {False: 'serve_str', True: 'serve_str_uds'},
}


class AbstractWorker:
    _idl = 'id'

    def __init__(self, parent: AbstractServer, idx: int, target: Any, args: Any):
        self.parent = parent
        self.idx = idx
        self.interrupt_by_parent = False
        self.birth = time.monotonic()
        self._spawn(target, args)

    def _spawn(self, target, args):
        raise NotImplementedError

    def _id(self):
        raise NotImplementedError

    def _watcher(self):
        self.inner.join()
        if not self.interrupt_by_parent:
            logger.error(f'Unexpected exit from worker-{self.idx + 1}')
            if self.parent.reload_on_changes and self.parent.reload_ignore_worker_failure:
                return
            self.parent.interrupt_children.append(self.idx)
            self.parent.main_loop_interrupt.set()

    def _watch(self):
        watcher = threading.Thread(target=self._watcher)
        watcher.start()

    def start(self):
        self.inner.start()
        logger.info(f'Spawning worker-{self.idx + 1} with {self._idl}: {self._id()}')
        self._watch()

    def is_alive(self):
        return self.inner.is_alive()

    def terminate(self):
        raise NotImplementedError

    def kill(self):
        raise NotImplementedError

    def join(self, timeout=None):
        self.inner.join(timeout=timeout)


class AbstractServer(Generic[WT]):
    def __init__(
        self,
        target: str,
        address: str = '127.0.0.1',
        port: int = 8000,
        uds: Path | None = None,
        uds_permissions: int | None = None,
        interface: Interfaces = Interfaces.RSGI,
        workers: int = 1,
        blocking_threads: int | None = None,
        blocking_threads_idle_timeout: int = 30,
        runtime_threads: int = 1,
        runtime_blocking_threads: int | None = None,
        runtime_mode: RuntimeModes = RuntimeModes.auto,
        loop: Loops = Loops.auto,
        task_impl: TaskImpl = TaskImpl.asyncio,
        http: HTTPModes = HTTPModes.auto,
        websockets: bool = True,
        backlog: int = 1024,
        backpressure: int | None = None,
        http1_settings: HTTP1Settings | None = None,
        http2_settings: HTTP2Settings | None = None,
        log_enabled: bool = True,
        log_level: LogLevels = LogLevels.info,
        log_dictconfig: dict[str, Any] | None = None,
        log_access: bool = False,
        log_access_format: str | None = None,
        ssl_cert: Path | None = None,
        ssl_key: Path | None = None,
        ssl_key_password: str | None = None,
        ssl_protocol_min: SSLProtocols = SSLProtocols.tls13,
        ssl_ca: Path | None = None,
        ssl_crl: list[Path] | None = None,
        ssl_client_verify: bool = False,
        url_path_prefix: str | None = None,
        respawn_failed_workers: bool = False,
        respawn_interval: float = 3.5,
        rss_sample_interval: int = 30,
        rss_samples: int = 1,
        workers_lifetime: int | None = None,
        workers_max_rss: int | None = None,
        workers_kill_timeout: int | None = None,
        factory: bool = False,
        working_dir: Path | None = None,
        env_files: Sequence[Path] | None = None,
        static_path_route: Sequence[str] | None = None,
        static_path_mount: Sequence[Path] | None = None,
        static_path_dir_to_file: str | None = None,
        static_path_expires: int = 86400,
        metrics_enabled: bool = False,
        metrics_scrape_interval: int = 15,
        metrics_address: str = '127.0.0.1',
        metrics_port: int = 9090,
        reload: bool = False,
        reload_paths: Sequence[Path] | None = None,
        reload_ignore_dirs: Sequence[str] | None = None,
        reload_ignore_patterns: Sequence[str] | None = None,
        reload_ignore_paths: Sequence[Path] | None = None,
        reload_filter: type[watchfiles.BaseFilter] | None = None,
        reload_tick: int = 50,
        reload_ignore_worker_failure: bool = False,
        process_name: str | None = None,
        pid_file: Path | None = None,
    ):
        self.target = target
        self.bind_addr = address
        self.bind_port = port
        self.bind_uds = uds.resolve() if uds else None
        self.uds_permissions = uds_permissions
        self.interface = interface
        self.workers = max(1, workers)
        self.runtime_threads = max(1, runtime_threads)
        self.runtime_blocking_threads = 512 if runtime_blocking_threads is None else max(1, runtime_blocking_threads)
        self.runtime_mode = runtime_mode
        self.loop = loop
        self.task_impl = task_impl
        self.http = http
        self.websockets = websockets
        self.backlog = max(128, backlog)
        self.backpressure = max(1, backpressure or self.backlog // self.workers)
        self.blocking_threads = (
            blocking_threads
            if blocking_threads is not None
            else max(1, (self.backpressure // 2) if self.interface == Interfaces.WSGI else 1)
        )
        self.blocking_threads_idle_timeout = blocking_threads_idle_timeout
        self.http1_settings = http1_settings
        self.http2_settings = http2_settings
        self.log_enabled = log_enabled
        self.log_level = log_level
        self.log_config = log_dictconfig
        self.log_access = log_access
        self.log_access_format = log_access_format or DEFAULT_ACCESSLOG_FMT
        self.url_path_prefix = url_path_prefix
        self.respawn_failed_workers = respawn_failed_workers
        self.reload_on_changes = reload
        self.respawn_interval = respawn_interval
        self.rss_sample_interval = rss_sample_interval
        self.rss_samples = rss_samples
        self._rss_wrk_samples = {}
        self.workers_lifetime = workers_lifetime
        self.workers_rss = workers_max_rss * 1024 * 1024 if workers_max_rss else None
        self.workers_kill_timeout = workers_kill_timeout
        self.factory = factory
        self.working_dir = working_dir
        self.env_files = env_files or ()
        self.static_path = None
        self.metrics_enabled = metrics_enabled
        self.metrics_scrape_interval = metrics_scrape_interval
        self.metrics_address = metrics_address
        self.metrics_port = metrics_port
        self.reload_paths = reload_paths or [Path.cwd()]
        self.reload_ignore_paths = reload_ignore_paths or ()
        self.reload_ignore_dirs = reload_ignore_dirs or ()
        self.reload_ignore_patterns = reload_ignore_patterns or ()
        self.reload_filter = reload_filter
        self.reload_tick = reload_tick
        self.reload_ignore_worker_failure = reload_ignore_worker_failure
        self.process_name = process_name
        self.pid_file = pid_file

        self.hooks_startup = []
        self.hooks_reload = []
        self.hooks_shutdown = []

        configure_logging(self.log_level, self.log_config, self.log_enabled)

        if static_path_mount:
            self._init_static_mounts(
                static_path_route or [],
                static_path_mount,
                static_path_dir_to_file,
                (str(static_path_expires) if static_path_expires else None),
            )
        self.build_ssl_context(
            ssl_cert, ssl_key, ssl_key_password, ssl_protocol_min, ssl_ca, ssl_crl or [], ssl_client_verify
        )
        self._ssp = None
        self._shd = None
        self._sfd = None
        self._metrics = MetricsAggregator(self.workers)
        self._metrics_exporter = MetricsExporter(self._metrics)
        self.wrks: list[WT] = []
        self.main_loop_interrupt = threading.Event()
        self.interrupt_signal = False
        self.interrupt_children = []
        self.respawned_wrks = {}
        self.reload_signal = False
        self.lifetime_signal = False
        self.rss_signal = False
        self.pid = None
        self._env_loader = build_env_loader()

    def _init_static_mounts(
        self,
        routes: Sequence[str],
        paths: Sequence[Path],
        dir_to_file: str | None,
        expires: str | None,
    ):
        if not paths:
            return
        if len(paths) == 1 and not routes:
            self.static_path = (
                [('/static', str(paths[0].resolve()))],
                dir_to_file,
                expires,
            )
            return
        if len(paths) != len(routes):
            logger.error('Static path routes and mounts should have the same length')
            raise ConfigurationError('static_path')
        self.static_path = (
            [(routes[idx], str(path.resolve())) for idx, path in enumerate(paths)],
            dir_to_file,
            expires,
        )

    def build_ssl_context(
        self,
        cert: Path | None,
        key: Path | None,
        password: str | None,
        proto: SSLProtocols,
        ca: Path | None,
        crl: list[Path],
        client_verify: bool,
    ):
        if not (cert and key):
            self.ssl_ctx = (False, None, None, None, str(proto), None, [], False)
            return
        # uneeded?
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ctx.load_cert_chain(str(cert.resolve()), str(key.resolve()), password)
        #: build ctx
        if client_verify and not ca:
            logger.warning('SSL client verification requires a CA certificate, ignoring')
            client_verify = False
        self.ssl_ctx = (
            True,
            str(cert.resolve()),
            str(key.resolve()),
            password,
            str(proto),
            str(ca.resolve()) if ca else None,
            [str(item.resolve()) for item in crl],
            client_verify,
        )

    @property
    def _bind_addr_fmt(self):
        return f'unix:{self.bind_uds}' if self.bind_uds else f'{self.bind_addr}:{self.bind_port}'

    @staticmethod
    def _call_hooks(hooks):
        for hook in hooks:
            hook()

    def on_startup(self, hook: Callable[[], Any]) -> Callable[[], Any]:
        self.hooks_startup.append(hook)
        return hook

    def on_reload(self, hook: Callable[[], Any]) -> Callable[[], Any]:
        self.hooks_reload.append(hook)
        return hook

    def on_shutdown(self, hook: Callable[[], Any]) -> Callable[[], Any]:
        self.hooks_shutdown.append(hook)
        return hook

    def _init_shared_socket(self):
        if self.bind_uds:
            self._ssp = UnixSocketSpec(str(self.bind_uds), self.backlog, self.uds_permissions)
        else:
            self._ssp = SocketSpec(self.bind_addr, self.bind_port, self.backlog)
        self._shd = self._ssp.build()
        self._sfd = self._shd.get_fd()

    def signal_handler_interrupt(self, *args, **kwargs):
        self.interrupt_signal = True
        self.main_loop_interrupt.set()

    def signal_handler_reload(self, *args, **kwargs):
        self.reload_signal = True
        self.main_loop_interrupt.set()

    def _spawn_worker(self, idx, target, callback_loader, socket_loader) -> WT:
        raise NotImplementedError

    def _spawn_workers(self, spawn_target, target_loader):
        for idx in range(self.workers):
            wrk = self._spawn_worker(idx=idx, target=spawn_target, callback_loader=target_loader)
            wrk.start()
            self.wrks.append(wrk)
        self._metrics.incr_spawn(self.workers)

    def _respawn_workers(self, workers, spawn_target, target_loader, delay: float = 0):
        for idx in workers:
            self.respawned_wrks[idx] = time.monotonic()
            logger.info(f'Respawning worker-{idx + 1}')
            old_wrk = self.wrks.pop(idx)
            wrk = self._spawn_worker(idx=idx, target=spawn_target, callback_loader=target_loader)
            wrk.start()
            self.wrks.insert(idx, wrk)
            time.sleep(delay)
            logger.info(f'Stopping old worker-{idx + 1}')
            old_wrk.terminate()
            old_wrk.join(self.workers_kill_timeout)
            if self.workers_kill_timeout:
                # the worker might still be reported alive after `join`, let's context switch
                if old_wrk.is_alive():
                    time.sleep(0.001)
                if old_wrk.is_alive():
                    logger.warning(f'Killing old worker-{idx + 1} after it refused to gracefully stop')
                    old_wrk.kill()
                    old_wrk.join()
        self._metrics.incr_spawn(len(workers))

    def _stop_workers(self):
        for wrk in self.wrks:
            wrk.terminate()

        for wrk in self.wrks:
            wrk.join(self.workers_kill_timeout)
            if self.workers_kill_timeout:
                # the worker might still be reported after `join`, let's context switch
                if wrk.is_alive():
                    time.sleep(0.001)
                if wrk.is_alive():
                    logger.warning(f'Killing worker-{wrk.idx} after it refused to gracefully stop')
                    wrk.kill()
                    wrk.join()

        self.wrks.clear()

    def _workers_lifetime_watcher(self, ttl):
        time.sleep(ttl)
        self.lifetime_signal = True
        self.main_loop_interrupt.set()

    def _workers_rss_watcher(self):
        time.sleep(self.rss_sample_interval)
        self.rss_signal = True
        self.main_loop_interrupt.set()

    def _watch_workers_lifetime(self, ttl):
        waker = threading.Thread(target=self._workers_lifetime_watcher, args=(ttl,), daemon=True)
        waker.start()

    def _watch_workers_rss(self):
        waker = threading.Thread(target=self._workers_rss_watcher, daemon=True)
        waker.start()

    def _write_pid(self):
        with self.pid_file.open('w') as pid_file:
            pid_file.write(str(self.pid))

    def _write_pidfile(self):
        if not self.pid_file:
            return

        existing_pid = None

        if self.pid_file.exists():
            try:
                with self.pid_file.open('r') as pid_file:
                    existing_pid = int(pid_file.read())
            except Exception:
                logger.error(f'Unable to read existing PID file {self.pid_file}')
                raise PidFileError

        if existing_pid is not None and existing_pid != self.pid:
            existing_process = True
            try:
                os.kill(existing_pid, 0)
            except OSError as e:
                if e.args[0] == errno.ESRCH:
                    existing_process = False

            if existing_process:
                logger.error(f'The PID file {self.pid_file} already exists for {existing_pid}')
                raise PidFileError

        self._write_pid()

    def _unlink_pidfile(self):
        if self.bind_uds and self.bind_uds.exists():
            self.bind_uds.unlink()

        if not (self.pid_file and self.pid_file.exists()):
            return

        try:
            with self.pid_file.open('r') as pid_file:
                file_pid = int(pid_file.read())
        except Exception:
            logger.error(f'Unable to read PID file {self.pid_file}')
            return

        if file_pid == self.pid:
            self.pid_file.unlink()

    def _start_ipc(self):
        pass

    def _stop_ipc(self):
        pass

    def _start_metrics(self):
        self._metrics_sig = WorkerSignal()
        self._metrics_exporter.run(
            SocketSpec(self.metrics_address, self.metrics_port, 128).build(),
            self._metrics_sig,
        )

    def _stop_metrics(self):
        self._metrics_sig.set()

    def startup(self, spawn_target, target_loader):
        self.pid = os.getpid()
        logger.info(f'Starting granian (main PID: {self.pid})')
        self._write_pidfile()
        set_main_signals(self.signal_handler_interrupt, self.signal_handler_reload)
        self._init_shared_socket()
        self._start_ipc()
        proto = 'https' if self.ssl_ctx[0] else 'http'
        logger.info(f'Listening at: {proto}://{self._bind_addr_fmt}')

        self._env_loader(self.env_files)
        self._call_hooks(self.hooks_startup)
        self._spawn_workers(spawn_target, target_loader)

        if self.workers_lifetime is not None:
            self._watch_workers_lifetime(self.workers_lifetime)
        if self.workers_rss is not None:
            self._watch_workers_rss()
        if self.metrics_enabled:
            self._start_metrics()

    def shutdown(self, exit_code=0):
        logger.info('Shutting down granian')
        if self.metrics_enabled:
            self._stop_metrics()
        self._stop_workers()
        self._stop_ipc()
        self._call_hooks(self.hooks_shutdown)
        self._unlink_pidfile()
        if not exit_code and self.interrupt_children:
            exit_code = 1
        if exit_code:
            sys.exit(exit_code)

    def _reload(self, spawn_target, target_loader):
        logger.info('HUP signal received, gracefully respawning workers..')
        workers = list(range(self.workers))
        self.reload_signal = False
        self.respawned_wrks.clear()
        self.main_loop_interrupt.clear()

        self._env_loader(self.env_files)
        self._call_hooks(self.hooks_reload)
        return self._respawn_workers(workers, spawn_target, target_loader, delay=self.respawn_interval)

    def _handle_rss_signal(self, spawn_target, target_loader):
        raise NotImplementedError

    def _serve_loop(self, spawn_target, target_loader):
        while True:
            self.main_loop_interrupt.wait()
            if self.interrupt_signal:
                break

            if self.interrupt_children:
                if not self.respawn_failed_workers:
                    break

                cycle = time.monotonic()
                if any(cycle - self.respawned_wrks.get(idx, 0) <= 5.5 for idx in self.interrupt_children):
                    logger.error('Worker crash loop detected, exiting')
                    break

                workers = list(self.interrupt_children)
                self.interrupt_children.clear()
                self.respawned_wrks.clear()
                self.main_loop_interrupt.clear()
                self._respawn_workers(workers, spawn_target, target_loader)
                self._metrics.incr_respawn_err(1)

            if self.reload_signal:
                self._reload(spawn_target, target_loader)

            if self.lifetime_signal or self.rss_signal:
                self.main_loop_interrupt.clear()

                if self.lifetime_signal:
                    self.lifetime_signal = False
                    ttl = self.workers_lifetime * 0.95
                    now = time.monotonic()
                    etas = [self.workers_lifetime]
                    for worker in list(self.wrks):
                        if (now - worker.birth) >= ttl:
                            logger.info(f'worker-{worker.idx + 1} lifetime expired, gracefully respawning..')
                            self._respawn_workers(
                                [worker.idx], spawn_target, target_loader, delay=self.respawn_interval
                            )
                            self._metrics.incr_respawn_ttl(1)
                        else:
                            elapsed = now - worker.birth
                            remaining = self.workers_lifetime - elapsed
                            etas.append(max(60, int(remaining)))
                    next_tick = min(etas)
                    self._watch_workers_lifetime(next_tick)

                if self.rss_signal:
                    self.rss_signal = False
                    self._handle_rss_signal(spawn_target, target_loader)
                    self._watch_workers_rss()

    def _serve(self, spawn_target, target_loader):
        self.startup(spawn_target, target_loader)
        self._serve_loop(spawn_target, target_loader)
        self.shutdown()

    def _serve_with_reloader(self, spawn_target, target_loader):
        if watchfiles is None:
            logger.error('Using --reload requires the granian[reload] extra')
            sys.exit(1)

        # Use given or default filter rules
        reload_filter_cls = self.reload_filter or watchfiles.filters.DefaultFilter
        # Extend `reload_filter` with provided args
        reload_filter_cls.ignore_dirs = (*reload_filter_cls.ignore_dirs, *self.reload_ignore_dirs)
        reload_filter_cls.ignore_entity_patterns = (
            *reload_filter_cls.ignore_entity_patterns,
            *self.reload_ignore_patterns,
        )
        reload_filter_cls.ignore_paths = (*reload_filter_cls.ignore_paths, *self.reload_ignore_paths)
        # Construct new filter
        reload_filter = reload_filter_cls()

        self.startup(spawn_target, target_loader)

        serve_loop = True
        while serve_loop:
            try:
                for changes in watchfiles.watch(
                    *self.reload_paths,
                    watch_filter=reload_filter,
                    stop_event=self.main_loop_interrupt,
                    step=self.reload_tick,
                ):
                    logger.info('Changes detected, reloading workers..')
                    for change, file in changes:
                        logger.info(f'{change.raw_str().capitalize()}: {file}')
                    self._env_loader(self.env_files)
                    self._call_hooks(self.hooks_reload)
                    self._stop_workers()
                    self._spawn_workers(spawn_target, target_loader)
            except StopIteration:
                pass

            if self.reload_signal:
                self._reload(spawn_target, target_loader)
            else:
                serve_loop = False

        self.shutdown()

    def serve(
        self,
        spawn_target: Callable[..., None] | None = None,
        target_loader: Callable[..., Callable[..., Any]] | None = None,
        wrap_loader: bool = True,
    ):
        default_spawners = {
            Interfaces.ASGI: self._spawn_asgi_lifespan_worker,
            Interfaces.ASGINL: self._spawn_asgi_worker,
            Interfaces.RSGI: self._spawn_rsgi_worker,
            Interfaces.WSGI: self._spawn_wsgi_worker,
        }
        if target_loader:
            if wrap_loader:
                target_loader = partial(target_loader, self.target)
        else:
            target_loader = partial(load_target, self.target, wd=self.working_dir, factory=self.factory)

        if not spawn_target:
            spawn_target = default_spawners[self.interface]
            if sys.platform == 'win32' and self.workers > 1:
                self.workers = 1
                logger.warn(
                    'Due to a bug in Windows unblocking socket implementation '
                    "granian can't support multiple workers on this platform. "
                    'Number of workers will now fallback to 1.'
                )

        if self.bind_uds and sys.platform == 'win32':
            logger.error('Unix Domain sockets are not available on Windows')
            raise ConfigurationError('uds')

        if self.interface != Interfaces.WSGI and self.blocking_threads > 1:
            logger.error('Blocking threads > 1 is not supported on ASGI and RSGI')
            raise ConfigurationError('blocking_threads')

        if self.websockets:
            if self.interface == Interfaces.WSGI:
                self.websockets = False
                logger.info('Websockets are not supported on WSGI, ignoring')
            if self.http == HTTPModes.http2:
                logger.info('Websockets are not supported on HTTP/2 only, ignoring')

        if setproctitle is not None:
            self.process_name = self.process_name or (f'granian {self.interface} {self._bind_addr_fmt} {self.target}')
            setproctitle.setproctitle(self.process_name)
        elif self.process_name is not None:
            logger.error('Setting process name requires the granian[pname] extra')
            raise ConfigurationError('process_name')

        if self.env_files and dotenv is None:
            logger.error('Environment file(s) usage requires the granian[dotenv] extra')
            raise ConfigurationError('env_files')

        if self.workers_lifetime is not None:
            if self.workers_lifetime < 60:
                logger.error('Workers lifetime cannot be less than 60 seconds')
                raise ConfigurationError('workers_lifetime')
            if self.reload_on_changes:
                self.workers_lifetime = None
                logger.info('Workers lifetime is not available in combination with changes reloader, ignoring')

        if self.workers_rss is not None:
            if self.reload_on_changes:
                self.workers_rss = None
                logger.info('The resource monitor is not available in combination with changes reloader, ignoring')

        if self.metrics_enabled:
            if self.reload_on_changes:
                self.metrics_enabled = False
                logger.info('Metrics are not available in combination with changes reloader, ignoring')

        if self.blocking_threads_idle_timeout < 5 or self.blocking_threads_idle_timeout > 600:
            logger.error('Blocking threads idle timeout must be between 5 and 600 seconds')
            raise ConfigurationError('blocking_threads_idle_timeout')

        cpus = multiprocessing.cpu_count()
        if self.workers > cpus:
            logger.warning(
                'Configured number of workers appears to be higher than the amount of CPU cores available. '
                'Mind that such value might actually decrease the overall throughput of the server. '
                f'Consider using {cpus} workers and tune threads configuration instead'
            )
        if self.runtime_threads > cpus:
            logger.warning(
                'Configured number of Rust threads appears to be too high given the amount of CPU cores available. '
                'Mind that Rust threads are not involved in Python code execution, and they almost never be the '
                'limiting factor in scaling. Consider configuring the amount of blocking threads instead'
            )

        if self.runtime_mode == RuntimeModes.auto:
            self.runtime_mode = RuntimeModes.st
            if any(
                [
                    self.interface != Interfaces.RSGI,
                    self.runtime_threads > 1,
                    self.http == HTTPModes.http2,
                ]
            ):
                self.runtime_mode = RuntimeModes.mt

        if self.task_impl == TaskImpl.rust:
            if _PYV >= _PY_312:
                self.task_impl = TaskImpl.asyncio
                logger.warning('Rust task implementation is not available on Python >= 3.12, falling back to asyncio')
            else:
                logger.warning('Rust task implementation is experimental!')

        serve_method = self._serve_with_reloader if self.reload_on_changes else self._serve
        serve_method(spawn_target, target_loader)



import os
import platform
from pathlib import Path

import httpx
import pytest


RANGE_FILE_CONTENT = '0123456789'


@pytest.mark.asyncio
@pytest.mark.skipif(bool(os.getenv('PGO_RUN')), reason='PGO build')
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_scope(rsgi_server, runtime_mode):
    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/info?test=true')

    assert res.status_code == 200
    assert res.headers['content-type'] == 'application/json'

    data = res.json()
    assert data['proto'] == 'http'
    assert data['http_version'] == '1.1'
    assert data['rsgi_version'] == '1.6'
    assert data['scheme'] == 'http'
    assert data['method'] == 'GET'
    assert data['path'] == '/info'
    assert data['query_string'] == 'test=true'
    assert data['headers']['host'] == f'localhost:{port}'
    assert not data['authority']


@pytest.mark.asyncio
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_body(rsgi_server, runtime_mode):
    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.post(f'http://localhost:{port}/echo', content='test')

    assert res.status_code == 200
    assert res.text == 'test'


@pytest.mark.asyncio
@pytest.mark.skipif(not bool(os.getenv('PGO_RUN')), reason='not PGO build')
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_body_large(rsgi_server, runtime_mode):
    data = ''.join([f'{idx}test'.zfill(8) for idx in range(0, 5000)])
    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.post(f'http://localhost:{port}/echo', content=data)

    assert res.status_code == 200
    assert res.text == data


@pytest.mark.asyncio
@pytest.mark.skipif(platform.python_implementation() == 'PyPy', reason='RSGI stream broken on PyPy')
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_body_stream_req(rsgi_server, runtime_mode):
    data = ''.join([f'{idx}test'.zfill(8) for idx in range(0, 5000)])
    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.post(f'http://localhost:{port}/echos', content=data)

    assert res.status_code == 200
    assert res.text == data


@pytest.mark.asyncio
@pytest.mark.skipif(platform.python_implementation() == 'PyPy', reason='RSGI stream broken on PyPy')
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_body_stream_res(rsgi_server, runtime_mode):
    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/stream')

    assert res.status_code == 200
    assert res.text == 'test' * 3


@pytest.mark.asyncio
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_file(asgi_server, runtime_mode):
    async with asgi_server(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/file')

    assert res.status_code == 200
    assert res.headers['content-type'] == 'image/png'
    assert res.headers['content-length'] == '95'


@pytest.mark.asyncio
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
@pytest.mark.parametrize('file_range', [(0, 10), (3, 8), (5, 6), (9, 10)])
async def test_file_range(rsgi_server, runtime_mode, file_range, tmp_path: Path):
    temp_file = tmp_path / 'temp_file.txt'
    with temp_file.open('w') as f:
        f.write(RANGE_FILE_CONTENT)

    start, end = file_range
    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.get(
            f'http://localhost:{port}/file_range',
            headers=[('file-path', str(temp_file)), ('range', f'bytes={start}-{end - 1}')],
        )
        assert res.status_code == 206
        assert res.headers['content-range'] == f'bytes {start}-{end - 1}/10'
        assert res.headers['content-length'] == f'{end - start}'
        assert res.text == RANGE_FILE_CONTENT[start:end]


@pytest.mark.asyncio
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_file_range_exceeding(rsgi_server, runtime_mode, tmp_path: Path):
    temp_file = tmp_path / 'temp_file.txt'
    with temp_file.open('w') as f:
        f.write(RANGE_FILE_CONTENT)

    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.get(
            f'http://localhost:{port}/file_range', headers=[('file-path', str(temp_file)), ('range', 'bytes=0-20')]
        )
        assert res.status_code == 206
        assert res.headers['content-range'] == 'bytes 0-9/10'
        assert res.headers['content-length'] == '10'
        assert res.text == RANGE_FILE_CONTENT


@pytest.mark.asyncio
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_file_range_unsatisfiable(rsgi_server, runtime_mode, tmp_path: Path):
    temp_file = tmp_path / 'temp_file.txt'
    with temp_file.open('w') as f:
        f.write(RANGE_FILE_CONTENT)

    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.get(
            f'http://localhost:{port}/file_range', headers=[('file-path', str(temp_file)), ('range', 'bytes=10-20')]
        )
        assert res.status_code == 416
        assert res.headers['content-range'] == 'bytes */10'
        assert res.headers['4xx-reason'] == 'out'

        res = httpx.get(
            f'http://localhost:{port}/file_range', headers=[('file-path', str(temp_file)), ('range', 'bytes=8-4')]
        )
        assert res.status_code == 416
        assert res.headers['content-range'] == 'bytes */10'
        assert res.headers['4xx-reason'] == 'invalid'


@pytest.mark.asyncio
@pytest.mark.skipif(bool(os.getenv('PGO_RUN')), reason='PGO build')
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_app_error(rsgi_server, runtime_mode):
    async with rsgi_server(runtime_mode, ws=False) as port:
        res = httpx.get(f'http://localhost:{port}/err_app')

    assert res.status_code == 500
